/*
 * gen-ioctl: Ioctl interface for generic ioctl calls
 * 
 * Copyright (C) 2005 TiVo Inc. All Rights Reserved.
 */

#include <linux/init.h>
#include <linux/mm.h>
#include <linux/fs.h>
#include <linux/module.h>
#include <asm/uaccess.h>
#include <linux/slab.h>
#include <linux/tivo-genioctl.h>
#include <linux/tivo-crashallassert.h>

#define GIOCTL_MAJOR 243

int gioctl_open (struct inode *inode, struct file *filp){
        return 0;
}

int gioctl_ioctl (struct inode *inode, struct file *file, 
                unsigned int cmd, unsigned long arg) {
        int err;

        if ( !arg ){
                printk(KERN_ERR "[gen_ioctl] Invalid arg\n");
                return -EINVAL;
        }
        if(_IOC_TYPE(cmd) != GIOCTL_IOC_MAGIC)
        {
                printk(KERN_ERR "[gen_ioctl] cmd (0x%x) mismatch, "
                                "called with 0x%x while expecting 0x%x\n", 
                                cmd, _IOC_TYPE(cmd), GIOCTL_IOC_MAGIC);
                return -EINVAL;
        }

        switch(cmd){
                case CRASHALLASSRT:
                        return tivocrashallassert(arg);
                break;
#ifdef CONFIG_TICK_KILLER
		case TICK_KILLER:
			return tick_killer_start(arg);
		break;
#endif
                case UPDATE_EXEC_CTX:
                {
                        struct k_arg list;
                        int strsz = sizeof(current->exec_cntxt) - 1;
                        int len ;
                        err = copy_from_user((void *)&list,(void *)arg,sizeof(struct k_arg));
                        if ( err || (list.len < 0) || (list.tag != TAG_EXEC_CTX) ) {
                                return -EFAULT;
                        }
                        if ( list.len == 0 ) {
                                strcpy(current->exec_cntxt,current->comm);
                                return 0;
                        }
                        
                        len = (list.len >= strsz) ? strsz:list.len;
                        err = copy_from_user((void *)current->exec_cntxt,(void *)list.ptr,len);
                        if ( err ) {
                                strcpy(current->exec_cntxt,current->comm);
                                return -EFAULT;
                        }
                        current->exec_cntxt[len] = '\0';
                        return 0;
                }
                break;
                default:
                        printk(KERN_ERR "gen_ioctl: Invalid command\n");
                        return -EINVAL;
        }
       return -EINVAL; 
}

static struct file_operations gioctl_fops = {
        ioctl:   gioctl_ioctl,
        open:    gioctl_open,
};

void __init init_gioctl_if(void){
        int result;
        result = register_chrdev(GIOCTL_MAJOR, "gioctl", &gioctl_fops);
        if ( result < 0 )
                printk(KERN_ERR "gioctl: "
                                "failed to register char dev [major %d]\n",GIOCTL_MAJOR);
}

module_init(init_gioctl_if);
